//-------------------------------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
//-------------------------------------------------------------------------------------------------
//----Parameters and limits for the marina
#define MAX_BOATS 125
#define SLIP_PRICE_PER_FOOT 12.50
#define LAND_PRICE_PER_FOOT 14.00
#define TRAILOR_PRICE_PER_FOOT 25.00
#define STORAGE_PRICE_PER_FOOT 11.20
//-------------------------------------------------------------------------------------------------
//----Data types
typedef int boolean;
#define TRUE 1
#define FALSE 0

#define STRING_LENGTH 128
typedef char String[STRING_LENGTH];

typedef enum {
    slip,
    land,
    trailor,
    storage,
    no_place
} PlaceType;

typedef union {
    unsigned short SlipNumber;
    char BayLetter;
    String LicenseTag;
    unsigned short StorageSpaceNumber;
} ExtraInformationUnion;

typedef struct {
    String Name;
    unsigned short Length;
    PlaceType Place;
    ExtraInformationUnion ExtraInformation;
    double AmountOwed;
} BoatInformation;

typedef BoatInformation* BoatInformationPtr;

typedef BoatInformationPtr AllBoatsInformation[MAX_BOATS];
//-------------------------------------------------------------------------------------------------
//----Safe generic malloc
void * Malloc(size_t Size) {

    void * Memory;

    if ((Memory = malloc(Size)) == NULL) {
        perror("Cannot malloc");
        exit(EXIT_FAILURE);
    } else {
        return(Memory);
    }
}
//-------------------------------------------------------------------------------------------------
//----Generic free that NULLs the pointer
void Free(void ** Memory) {

    free(*Memory);
    *Memory = NULL;
}
//-------------------------------------------------------------------------------------------------
//----Convert a string to a place
PlaceType StringToPlaceType(char * PlaceString) {

    if (!strcasecmp(PlaceString,"slip")) {
        return(slip);
    }
    if (!strcasecmp(PlaceString,"land")) {
        return(land);
    }
    if (!strcasecmp(PlaceString,"trailor")) {
        return(trailor);
    }
    if (!strcasecmp(PlaceString,"storage")) {
        return(storage);
    }
    return(no_place);
}
//-------------------------------------------------------------------------------------------------
//----Convert a place to a string
char * PlaceToString(PlaceType Place) {

    switch (Place) {
        case slip:
            return("slip");
        case land:
            return("land");
        case trailor:
            return("trailor");
        case storage:
            return("storage");
        case no_place:
            return("no_place");
        default:
            printf("How the faaark did I get here?\n");
            exit(EXIT_FAILURE);
            break;
    }
}
//-------------------------------------------------------------------------------------------------
//----Free all boat data. Assumes they are packed.
void FreeBoatInformation(AllBoatsInformation TheBoats) {

    int BoatIndex = 0;

    while (TheBoats[BoatIndex] != NULL) {
        Free((void **)&(TheBoats[BoatIndex]));
        BoatIndex++;
    }
}
//-------------------------------------------------------------------------------------------------
//----Convert a CSV line into a boat struct
BoatInformationPtr CSVLineToBoatInformation(char * CSVLine,char * CSVFileName) {

    BoatInformationPtr NewBoat;
    int FieldIndex;
    char * FieldPtr;
    boolean Error = FALSE;

//----Get memory for the struct
    NewBoat = (BoatInformationPtr)Malloc(sizeof(BoatInformation));
//----Break up the CSV line at commas
    FieldPtr = strtok(CSVLine,",");
//----Convert each of the fields into the right data type in the struct
    for (FieldIndex = 1;!Error && FieldIndex <= 5;FieldIndex++) {
        switch (FieldIndex) {
//----1st field is the name - copy the string
            case 1:
                strcpy(NewBoat->Name,FieldPtr);
//DEBUG printf("Boat name is %s\n",NewBoat->Name);
                break;
            case 2:
//----2nd field is the length - convert the string to an integer
                NewBoat->Length = atoi(FieldPtr);
//DEBUG printf("Boat length is %d\n",NewBoat->Length);
                break;
            case 3:
//----3rd field is the place - convert the string to a place
                NewBoat->Place = StringToPlaceType(FieldPtr);
//DEBUG printf("Boat place is %s\n",PlaceToString(NewBoat->Place));
                break;
//----4th field depends on the place
            case 4:
                switch (NewBoat->Place) {
//----4th field is a slip - convert the string to an integer slip number
                    case slip:
                        NewBoat->ExtraInformation.SlipNumber = atoi(FieldPtr);
//DEBUG printf("Boat slip number is %d\n",NewBoat->ExtraInformation.SlipNumber);
                        break;
                    case land:
//----4th field is land - use the first character as the bay letter
                        NewBoat->ExtraInformation.BayLetter = FieldPtr[0];
//DEBUG printf("Boat bay letter is %c\n",NewBoat->ExtraInformation.BayLetter);
                        break;
                    case trailor:
//----4th field is a trailer - copy the string to the license tag
                        strcpy(NewBoat->ExtraInformation.LicenseTag,FieldPtr);
//DEBUG printf("Boat trailor tag is %s\n",NewBoat->ExtraInformation.LicenseTag);
                        break;
//----4th field is storage - convert the string to an integer storage space number
                    case storage:
                        NewBoat->ExtraInformation.StorageSpaceNumber = atoi(FieldPtr);
//DEBUG printf("Boat storage number is %d\n",NewBoat->ExtraInformation.StorageSpaceNumber);
                        break;
//----Otherwise something is wrong. Return the memory and return NULL for an error
                    default:
                        printf("Error in CSV file %s\n",CSVFileName);
                        Free((void *)&NewBoat);
                        return(NULL);
                        break;
                }
                break;
            case 5:
//----5th field is the amount owed - convert the string to a float
                NewBoat->AmountOwed = atof(FieldPtr);
//DEBUG printf("Boat owes $%.2f\n",NewBoat->AmountOwed);
                break;
//----This should be impossible
            default:
                printf("How the faaark did I get here?\n");
                exit(EXIT_FAILURE);
                break;
        }
//----Get next CSV field
        FieldPtr = strtok(NULL,",");
    }

//----Return pointer to the struct
    return(NewBoat);
}
//-------------------------------------------------------------------------------------------------
//----Load the boat information from a CSV file
boolean LoadBoatInformation(char * CSVFileName,AllBoatsInformation TheBoats) {

    FILE * BoatFile;
    String CSVLine;
    int BoatIndex = 0;

//----Open the file for reading
    if ((BoatFile = fopen(CSVFileName,"r")) == NULL) {
        printf("Could not open CSV file %s for reading\n",CSVFileName);
        return(FALSE);
    }

//----Read file line by line, and convert each to a boat struct
    while (fgets(CSVLine,STRING_LENGTH,BoatFile) != NULL) {
        CSVLine[strlen(CSVLine)-1] = '\0';
//DEBUG printf("Line is %s\n",CSVLine);
        if ((TheBoats[BoatIndex] = CSVLineToBoatInformation(CSVLine,CSVFileName)) != NULL) {
            BoatIndex++;
        }
    }

//----Set the rest to NULL
    while (BoatIndex < MAX_BOATS) {
        TheBoats[BoatIndex++] = NULL;
    }

//----Close the file
    if (fclose(BoatFile) != 0) {
        printf("Could not close CSV file %s after reading\n",CSVFileName);
        return(FALSE);
    }

    return(TRUE);
}
//-------------------------------------------------------------------------------------------------
//----Save the boat information to a CSV file
boolean SaveBoatInformation(char * CSVFileName,AllBoatsInformation TheBoats) {

    FILE * BoatFile;
    int BoatIndex;

//----Open the file for writing
    if ((BoatFile = fopen(CSVFileName,"w")) == NULL) {
        printf("Could not open CSV file %s for writing\n",CSVFileName);
        return(FALSE);
    }

//----Loop until NULL is reached - assumes array is packed
    BoatIndex = 0;
    while(TheBoats[BoatIndex] != NULL) {
//----Print first three fields - name,length,place, - directly
        fprintf(BoatFile,"%s,%d,%s,",TheBoats[BoatIndex]->Name,TheBoats[BoatIndex]->Length,
PlaceToString(TheBoats[BoatIndex]->Place));
//----Print the place field according to its type
        switch (TheBoats[BoatIndex]->Place) {
            case slip:
                fprintf(BoatFile,"%d",TheBoats[BoatIndex]->ExtraInformation.SlipNumber);
                break;
            case land:
                fprintf(BoatFile,"%c",TheBoats[BoatIndex]->ExtraInformation.BayLetter);
                break;
            case trailor:
                fprintf(BoatFile,"%s",TheBoats[BoatIndex]->ExtraInformation.LicenseTag);
                break;
            case storage:
                fprintf(BoatFile,"%d",TheBoats[BoatIndex]->ExtraInformation.StorageSpaceNumber);
                break;
            case no_place:
            default:
                printf("How the faaark did I get here?\n");
                exit(EXIT_FAILURE);
                break;
        }
//----Print the amount owed directly
        fprintf(BoatFile,",%.2f\n",TheBoats[BoatIndex]->AmountOwed);
        BoatIndex++;
    }

//----Close the file
    if (fclose(BoatFile) != 0) {
        printf("Could not close CSV file %s after writing\n",CSVFileName);
        return(FALSE);
    }

    return(TRUE);
}
//-------------------------------------------------------------------------------------------------
//----Compare pointer to a pointer to a boat struct by the names in the struct
int CompareBoats(const void * First,const void * Second) {

//DEBUG printf("Compare %s with %s is %d\n",(*((BoatInformationPtr *)First))->Name,
//DEBUG (*((BoatInformationPtr *)Second))->Name,strcmp((*((BoatInformationPtr *)First))->Name,
//DEBUG (*((BoatInformationPtr *)Second))->Name));
    return(strcasecmp((*((BoatInformationPtr *)First))->Name,
(*((BoatInformationPtr *)Second))->Name));
}
//-------------------------------------------------------------------------------------------------
//----Sort the pointers to the structs by boat name, packing first, NULLifying the rest
void SortBoatsByName(AllBoatsInformation TheBoats) {

    int TakeFromIndex;
    int InsertToIndex = 0;
    int NumberOfBoats;

//----Pack the boats up tight, removing NULL pointers
    for (TakeFromIndex = 0;TakeFromIndex < MAX_BOATS;TakeFromIndex++) {
        if (TheBoats[TakeFromIndex] != NULL) {
            TheBoats[InsertToIndex] = TheBoats[TakeFromIndex];
            InsertToIndex++;
        }
    }
    NumberOfBoats = InsertToIndex;
//----Make sure the rest are NULL
    while (InsertToIndex < MAX_BOATS) {
        TheBoats[InsertToIndex++] = NULL;
    }

//----Sort the pointers by name in struct
    qsort((void *)TheBoats,NumberOfBoats,sizeof(BoatInformationPtr),CompareBoats);
}
//-------------------------------------------------------------------------------------------------
//----Search for a struct by boat name, case insensitive. Assumes the array is packed.
int FindBoatWithName(char * BoatName,AllBoatsInformation TheBoats) {

    int BoatIndex = 0;

//----Loop until end of array, or no more boats, or boat found (note short circuit evaluation)
    while (BoatIndex < MAX_BOATS && TheBoats[BoatIndex] != NULL && 
strcasecmp(BoatName,TheBoats[BoatIndex]->Name)) {
        BoatIndex++;
    }
//----If found return the index, else -1 for not found
    if (BoatIndex >= 0 && TheBoats[BoatIndex] != NULL) {
        return(BoatIndex);
    } else {
        return(-1);
    }
}
//-------------------------------------------------------------------------------------------------
//----Search for a struct with a boat name input by user
int FindBoatByName(AllBoatsInformation TheBoats) {

    String BoatName;
    int BoatIndex;

//----Get the boat name
    printf("Please enter the boat name                               : ");
    fgets(BoatName,STRING_LENGTH,stdin);
    BoatName[strlen(BoatName)-1] = '\0';
//----Check if found in storage
    if ((BoatIndex = FindBoatWithName(BoatName,TheBoats)) == -1) {
        printf("No boat with that name\n");
    }
    return(BoatIndex);
}
//-------------------------------------------------------------------------------------------------
//----Print the boat inventory. Assumes array is packed.
void PrintInventory(AllBoatsInformation TheBoats) {

    int BoatIndex = 0;

//----Loop through boats
    while (TheBoats[BoatIndex] != NULL) {
//----Print first three fields - name,length,place, - directly
        printf("%-20s %3d' %7s ",TheBoats[BoatIndex]->Name,TheBoats[BoatIndex]->Length,
PlaceToString(TheBoats[BoatIndex]->Place));
//----Print the place field according to its type
        switch (TheBoats[BoatIndex]->Place) {
            case slip:
                printf("  #%3d",TheBoats[BoatIndex]->ExtraInformation.SlipNumber);
                break;
            case land:
                printf("     %c",TheBoats[BoatIndex]->ExtraInformation.BayLetter);
                break;
            case trailor:
                printf("%-6s",TheBoats[BoatIndex]->ExtraInformation.LicenseTag);
                break;
            case storage:
                printf("  #%3d",TheBoats[BoatIndex]->ExtraInformation.StorageSpaceNumber);
                break;
            case no_place:
            default:
                printf("How the faaark did I get here?\n");
                exit(EXIT_FAILURE);
                break;
        }
//----Print the amount owed directly
        printf("   Owes $%7.2f\n",TheBoats[BoatIndex]->AmountOwed);
//----Move to next boat
    BoatIndex++;
    }
}
//-------------------------------------------------------------------------------------------------
//----Add a boat to the storage data
void AddBoat(AllBoatsInformation TheBoats) {

    String CSVLine;
    int BoatIndex;

//----Look for an unused place in storage record
    BoatIndex = 0;
    while (BoatIndex < MAX_BOATS && TheBoats[BoatIndex] != NULL) {
        BoatIndex++;
    }
//----Check if unused place has been found
    if (BoatIndex == MAX_BOATS) {
        printf("No space for another boat\n");
        return;
    }
//----Get the new boat data in CSV format
    printf("Please enter the boat data in CSV format                 : ");
    fgets(CSVLine,STRING_LENGTH,stdin);
    CSVLine[strlen(CSVLine)-1] = '\0';
//----Add storage record for that data
    TheBoats[BoatIndex] = CSVLineToBoatInformation(CSVLine,"user");
//----Resort the data
    SortBoatsByName(TheBoats);
}
//-------------------------------------------------------------------------------------------------
//----Remove a boat from the storage data
void RemoveBoat(AllBoatsInformation TheBoats) {

    int BoatIndex;

//----Get boat index in storage by name
    if ((BoatIndex = FindBoatByName(TheBoats)) == -1) {
        return;
    }
//----Release the storage record
    Free((void *)(&TheBoats[BoatIndex]));
//----Resort the data
    SortBoatsByName(TheBoats);
}
//-------------------------------------------------------------------------------------------------
//----Make a payment up to the amount owed
void PayForBoat(AllBoatsInformation TheBoats) {

    int BoatIndex;
    double PaymentAmount;

//----Get boat index in storage by name
    if ((BoatIndex = FindBoatByName(TheBoats)) == -1) {
        return;
    }
//----Get payment amount
    printf("Please enter the amount to be paid                       : ");
    scanf(" %lf",&PaymentAmount);
    if (PaymentAmount > TheBoats[BoatIndex]->AmountOwed) {
        printf("That is more than the amount owed, $%.2f\n",TheBoats[BoatIndex]->AmountOwed);
        return;
    }
    TheBoats[BoatIndex]->AmountOwed -= PaymentAmount;
}
//-------------------------------------------------------------------------------------------------
//----Add monthly charge to all boats. Assumes array is packed.
void AddMonthPayment(AllBoatsInformation TheBoats) {

    int BoatIndex = 0;
    double PricePerFoot;

//----Loop through the boats
    while (TheBoats[BoatIndex] != NULL) {
//----Select rate according to place
        switch (TheBoats[BoatIndex]->Place) {
            case slip:
                PricePerFoot = SLIP_PRICE_PER_FOOT;
                break;
            case land:
                PricePerFoot = LAND_PRICE_PER_FOOT;
                break;
            case trailor:
                PricePerFoot = TRAILOR_PRICE_PER_FOOT;
                break;
            case storage:
                PricePerFoot = STORAGE_PRICE_PER_FOOT;
                break;
            case no_place:
            default:
                printf("How the faaark did I get here?\n");
                exit(EXIT_FAILURE);
                break;
        }
//----Add on monthly amount according to length
        TheBoats[BoatIndex]->AmountOwed += TheBoats[BoatIndex]->Length * PricePerFoot;
//----Move to next boat
        BoatIndex++;
    }

}
//-------------------------------------------------------------------------------------------------
//----Use menu
void RunMenu(AllBoatsInformation TheBoats) {

    char UserInput;

//----Welcome the user
    printf("\nWelcome to the Boat Management System\n");
    printf("-------------------------------------\n");
    do {
//----Print options
        printf("\n(I)nventory, (A)dd, (R)emove, (P)ayment, (M)onth, e(X)it : ");
//----Read and convert to uppercase. Note the %*c to eat the \n
        scanf(" %c%*c",&UserInput);
        UserInput = toupper(UserInput);
//----Perform select action
        switch (UserInput) {
            case 'I':
                PrintInventory(TheBoats);
                break;
            case 'A':
                AddBoat(TheBoats);
                break;
            case 'R':
                RemoveBoat(TheBoats);
                break;
            case 'P':
                PayForBoat(TheBoats);
                break;
            case 'M':
                AddMonthPayment(TheBoats);
                break;
            case 'X':
                printf("\nExiting the Boat Management System\n");
                break;
            default:
                printf("Invalid option %c\n",UserInput);
                break;
        }
//----Until exit
    } while (UserInput != 'X');
}
//-------------------------------------------------------------------------------------------------
int main(int argc,char *argv[]) {

    AllBoatsInformation AllBoats;

//----Check that there are enough arguments
    if (argc != 2) {
        printf("Usage: %s <CSV filename>\n",argv[0]);
        exit(EXIT_FAILURE);
    }

//----Load from the CSV file
    if (!LoadBoatInformation(argv[1],AllBoats)) {
        exit(EXIT_FAILURE);
    }

//----Sort the boats. This ensures they are packed, and the rest NULLified
    SortBoatsByName(AllBoats);
//----Run the menu
    RunMenu(AllBoats);

//----Save to the CSV file
    if (!SaveBoatInformation(argv[1],AllBoats)) {
        exit(EXIT_FAILURE);
    }
//----Free all dynamicaly allocated memory
    FreeBoatInformation(AllBoats);
}
//-------------------------------------------------------------------------------------------------
